from __future__ import annotations

__all__ = ["Hand"]

import ctypes
import sys
from collections.abc import Iterable, Iterator
from typing import Union

import endplay._dds as _dds
from endplay.config import suppress_unicode
from endplay.types.card import Card
from endplay.types.denom import Denom
from endplay.types.rank import Rank
from endplay.types.suitholding import SuitHolding


class Hand:
	"Class allowing manipulations of cards in the hand of a single player"
	def __init__(self, data: Union[str, ctypes.Array[ctypes.c_uint]] = "..."):
		"""
		Construct a hand object

		:param data: Either a PBN string of the hand, or a reference to a _dds object
		"""
		if isinstance(data, str):
			self._data = (ctypes.c_uint * 4)(0, 0, 0, 0)
			suits = data.split(".")
			if len(suits) > 4:
				raise RuntimeError(f"hand contains more than four suits: {data})")
			for i, suit in enumerate(suits):
				if suit and suit[0].upper() in "SHDC":
					suit = suit[1:]
				for rank in suit:
					self._data[i] |= Rank.find(rank).value
		else:
			self._data = data

	def __copy__(self) -> 'Hand':
		return Hand((ctypes.c_uint * 4).from_buffer_copy(self._data))

	def copy(self) -> 'Hand':
		return self.__copy__()

	def add(self, card: Card) -> bool:
		"""
		Adds a card to the hand

		:param card: The card to be added to the hand
		:return: False if the card was already in the hand, True otherwise
		"""
		if isinstance(card, str):
			card = Card(name = card)
		elif not isinstance(card, Card):
			raise ValueError("card must be of type Card or str")

		if card in self:
			return False
		self._data[card.suit.value] |= card.rank.value
		return True

	def extend(self, cards: Iterable[Card]) -> int:
		"""
		Add multiple cards to the hand

		:param cards: An iterable of the cards to add
		:return: The number of cards successfully added
		"""
		return sum(self.add(card) for card in cards)

	def remove(self, card: Card) -> bool:
		"""
		Remove a card from the hand

		:param card: The card to be added to the hand, can be a string
			representation e.g. "CQ"
		:return: False if the card wasn't in the hand, True otherwise
		"""
		if isinstance(card, str):
			card = Card(name = card)
		elif not isinstance(card, Card):
			raise ValueError("card must be of type Card or str")
		if card in self:
			self._data[card.suit.value] &= ~card.rank.value
			return True
		return False

	@staticmethod
	def from_pbn(pbn: str) -> 'Hand':
		"""
		Construct a Hand from a PBN string

		:param pbn: A PBN string for a hand, e.g. "QT62..AQT852.QJT"
		"""
		return Hand(pbn)

	def to_pbn(self) -> str:
		":return: A PBN representation of the hand"
		return '.'.join(str(self[suit]) for suit in Denom.suits())

	@staticmethod
	def from_lin(lin: str) -> 'Hand':
		"""
		Construct a Hand from a LIN string

		:param lin: A LIN string for a hand, e.g. "SQT62HDAQT852CQJT"
		"""
		pbn = lin[1:].replace("H", ".").replace("D", ".").replace("C", ".")
		return Hand.from_pbn(pbn)

	def to_lin(self) -> str:
		"""
		Convert a Hand to a LIN string
		"""
		lin = ""
		with suppress_unicode():
			for suit in Denom.suits():
				lin += suit.abbr + self[suit].to_pbn()[::-1]
		return lin


	def to_LaTeX(self, vertical: bool = True, ten_as_letter: bool = False) -> str:
		"""
		Create a LaTeX representation of the hand.

		:param vertical: If True uses \\vhand, else \\hhand layout
		:param title: The hand title. If vertical is False this is ignored
		"""
		if vertical:
			res, sep = r"\begin{tabular}{l}", r"\\"
		else:
			res, sep = "", " "
		for suit in Denom.suits():
			res += "$\\" + suit.name + "uit$ "
			if len(self[suit]) == 0:
				res += "---"
			else:
				for rank in self[suit]:
					res += r"\makebox[.8em]{"
					res += r"1\kern-.16em0" if (rank == Rank.RT and not ten_as_letter) else rank.abbr
					res += "}"
			res += sep
		if vertical:
			res += r"\end{tabular}"
		return res

	def pprint(self, vertical: bool = True, stream=sys.stdout) -> None:
		"Print the suits in the hand using suit symbols"
		sep = "\n" if vertical else " "
		print(sep.join(f"{suit.abbr} {self[suit]}" for suit in Denom.suits()), file=stream)

	def clear(self) -> None:
		"Remove all cards from the hand"
		for i in range(4):
			self._data[i] &= 0

	@property
	def spades(self) -> SuitHolding:
		"The spade holding of the hand"
		return self[Denom.spades]
	@spades.setter
	def spades(self, suit: SuitHolding) -> None:
		self[Denom.spades] = suit

	@property
	def hearts(self) -> SuitHolding:
		"The heart holding of the hand"
		return self[Denom.hearts]
	@hearts.setter
	def hearts(self, suit: SuitHolding) -> None:
		self[Denom.hearts] = suit

	@property
	def diamonds(self) -> SuitHolding:
		"The diamond holding of the hand"
		return self[Denom.diamonds]
	@diamonds.setter
	def diamonds(self, suit: SuitHolding) -> None:
		self[Denom.diamonds] = suit

	@property
	def clubs(self) -> SuitHolding:
		"The club holding of the hand"
		return self[Denom.clubs]
	@clubs.setter
	def clubs(self, suit: SuitHolding) -> None:
		self[Denom.clubs] = suit

	def __contains__(self, card: Card) -> bool:
		":return: True if card is in this hand"
		if isinstance(card, str):
			card = Card(name = card)
		elif not isinstance(card, Card):
			raise ValueError("card must be of type Card or str")
		return bool(self._data[card.suit.value] & card.rank.value)

	def __str__(self) -> str:
		":return: A PBN string representation of the hand"
		return self.to_pbn()

	def __repr__(self) -> str:
		return f'Hand("{self.to_pbn()}")'

	def __iter__(self) -> Iterator[Card]:
		":return: An iterator over the suit holdings in the order spades, hearts, diamonds and clubs"
		for suit in Denom.suits():
			for rank in self[suit]:
				yield Card(suit=suit, rank=rank)

	def __getitem__(self, suit: Denom) -> SuitHolding:
		":return: The specified suit holding of the hand"
		return SuitHolding(self._data, suit.value)

	def __setitem__(self, suit: Denom, holding: SuitHolding) -> None:
		if isinstance(holding, str):
			self[suit].clear()
			for rank in holding:
				self[suit].add(rank)
		else:
			self._data[suit] = holding._data[holding._idx]

	def __len__(self) -> int:
		":return: The number of cards in the hand"
		return sum(bin(suit).count('1') for suit in self._data)

	def __eq__(self, other: object) -> bool:
		if not isinstance(other, Hand):
			return NotImplemented
		return not _dds._libc.memcmp(self._data, other._data, len(self._data))



"""
A tkinter interface to many of the features of endplay including 
dealing hands and performing double-dummy analysis.
"""

__all__ = ["TkFrontend"]

from endplay.interact import InteractiveDeal

class TkFrontend:
    def __init__(self, deal: InteractiveDeal):
        raise NotImplementedError


"""
The Actions classes provide a common interface for producing different
types of output from a dealer script. When a script is run, an appropriate
Actions object is constructed, and any time output is requested one of the
methods is called to format it correctly.
"""

__all__ = ["BaseActions", "TerminalActions", "LaTeXActions", "HTMLActions"]

from endplay.dealer.actions.base import BaseActions
from endplay.dealer.actions.terminal import TerminalActions
from endplay.dealer.actions.latex import LaTeXActions
from endplay.dealer.actions.html import HTMLActions



"""
Endplay - A bridge tools library with generating, analysing and scoring.
Released under the MIT licence (see the LICENCE file provided with this distribution)
"""

import endplay._dds as _dds
from endplay.dds import *
from endplay.dealer import *
from endplay.interact import *
from endplay.parsers import *
from endplay.evaluate import *
from endplay.types import *
from endplay.config import \
	__version__, __version_info__, __author__, \
	__buildtime__, suppress_unicode


from __future__ import annotations

__all__ = ["Board"]

from endplay.types.deal import Deal
from endplay.types.contract import Contract
from endplay.types.vul import Vul
from endplay.types.player import Player
from endplay.types.bid import Bid
from endplay.types.card import Card
from typing import Optional, Any
from collections.abc import Iterable

class Board:
	"""
	Class representing a deal along with the play, auction and other table
	information

	:ivar deal: The deal at the table
	:vartype deal: Deal
	:ivar auction: The auction at the table
	:vartype auction: list[Bid]
	:ivar contract: The contract at the table
	:vartype contract: Contract
	:ivar play: The play history at the table
	:vartype play: list[Card]
	:ivar board_num: The board number of this deal
	:vartype board_num: int
	:ivar vul: The board vulnerability. If this isn't defined
		(i.e. set to `None`) then it is deduced from `board_num`
	:vartype vul: Vul
	:ivar dealer: The board dealer. Similarly to `vul` this can
		be deduced from `board_num`
	:vartype dealer: Player
	:ivar claimed: Flag indicating whether the play ended as a result
		of a claim
	:vartype claimed: bool
	:ivar info: A dictionary which contains arbitrary extra information
		about the board. The dictionary type used provided case-insensitive
		dot-access as a convenience (i.e. `board.info.event` and `board.info.Event`
		refer to the same object, but `board.info["event"]` and `board.info["Event"]`
		would be considered different). Tabular data can be stored here, any
		key ending with (but not equal to) `table` is treated as a table and its value
		should be a dictionary containing two keys: `headers` with a list of column names,
		and `rows` with a list of the rows. The column names can either be plain strings,
		or dictionaries with the keys

		* `ordering`: Either `"+"`, `"-"` or `None` depending of if the table is sorted
			ascending, descending or unsorted with respect to this column
		* `name`: A string value with the name of the column
		* `minwidth`: The minimum width that values in this column should be
		* `alignment`: `"L"` or `"R"` depending on if this column should be left or right
			aligned. Ignored unless `minwidth` is defined
	"""
	class Info(dict):
		"""
		Dictionary-like class which alows for case-insensitive dot-access,
		for example::

			info["Event"] = "WBF 2017"
			print(info.event) # WBF 2017
		"""
		def _find_key(self, key: str):
			key = key.casefold()
			for k in self:
				if k.casefold() == key:
					return k
			return None
		def __getattr__(self, attr: str) -> Any:
			key = self._find_key(attr)
			if key is not None:
				return self[key]
			return None
		def __setattr__(self, attr: str, value: Any) -> None:
			key = self._find_key(attr)
			if key is not None:
				self[key] = value
			else:
				self[attr] = value
		def __delattr__(self, attr: str) -> None:
			key = self._find_key(attr)
			if key is not None:
				del self[key]
			else:
				raise KeyError(attr)

	def __init__(self, 
		deal: Optional[Deal] = None,
		auction: Optional[Iterable[Bid]] = None,
		play: Optional[Iterable[Card]] = None,
		board_num: Optional[int] = None,
		*,
		vul: Optional[Vul] = None,
		dealer: Optional[Player] = None,
		contract: Optional[Contract] = None,
		claimed: bool = False,
		**kwargs):
		self.deal = deal.copy() if deal is not None else Deal()
		self.auction = list(auction) if auction is not None else []
		self.play = list(play) if play is not None else []
		self.board_num = board_num
		self._dealer = dealer
		self._vul = vul
		self._contract = contract
		self.claimed = claimed
		self.info = Board.Info(**kwargs)

	@property
	def dealer(self) -> Optional[Player]:
		"""
		Dealer of the board. If not defined, then attempts to
		calculate based on the value of `board_num`
		"""
		if self._dealer is not None:
			return self._dealer
		elif self.board_num is not None:
			return Player.from_board(self.board_num)
		else:
			return None
	@dealer.setter
	def dealer(self, value: Player) -> None:
		self._dealer = value

	@property
	def vul(self) -> Optional[Vul]:
		"""
		Vulnerability of the board. If not defined, then attempts
		to calculate based on the value of `board_num`
		"""
		if self._vul is not None:
			return self._vul
		elif self.board_num is not None:
			return Vul.from_board(self.board_num)
		else:
			return None
	@vul.setter
	def vul(self, value: Vul) -> None:
		self._vul = value

	@property
	def contract(self) -> Optional[Contract]:
		"""
		The contract the board was played in. If not provided, then
		attempts to calculate based on the auction and play history.
		"""
		if self._contract is not None:
			return self._contract
		elif self.auction and self.dealer:
			c = Contract.from_auction(self.dealer, self.auction)
			if self.play:
				from endplay.utils.play import total_tricks, tricks_to_result
				c.result = tricks_to_result(
					total_tricks(self.play, self.deal.trump), 
					c.level)
			return c
		else:
			return None
	@contract.setter
	def contract(self, value: Contract) -> None:
		self._contract = value



"""
Base actions class to provide the interface.
"""

__all__ = ["BaseActions"]

import sys
from importlib.resources import read_text
from typing import Optional, Union
from abc import ABC, abstractmethod
from endplay.types import Player
from endplay.dealer.constraint import Expr

class BaseActions(ABC):
	def __init__(self, deals, stream, board_numbers, template_ext=None):
		self.board_numbers = board_numbers
		self.deals = deals
		self.stream = sys.stdout if stream is None else stream
		self.write = lambda *objs, **kwargs: print(*objs, **kwargs, file=self.stream)
		if template_ext is not None:
			self.preamble = read_text("endplay.dealer.actions.templates", "preamble." + template_ext)
			self.postamble = read_text("endplay.dealer.actions.templates", "postamble." + template_ext)
		else:
			self.preamble = self.postamble = ""

	def write(self, *objs, **kwargs):
		if "file" in kwargs:
			raise RuntimeError("Unexpected keyword argument 'file' passed to BaseActions.write")
		print(*objs, **kwargs, file=self.stream)

	def write_preamble(self):
		self.write(self.preamble)

	def write_postamble(self):
		self.write(self.postamble)

	def printall(self):
		self.print(*Player)

	@abstractmethod
	def print(self, *players):
		raise NotImplementedError

	def printew(self):
		self.print(Player.east, Player.west)

	@abstractmethod
	def printpbn(self):
		pass

	@abstractmethod
	def printcompact(self, expr: Expr = None):
		pass

	@abstractmethod
	def printoneline(self, expr: Expr = None):
		pass

	@abstractmethod
	def printes(self, *objs: Union[Expr, str]):
		pass

	@abstractmethod
	def average(self, expr: Expr, s: Optional[str] = None):
		pass

	@abstractmethod
	def frequency1d(
		self, expr: Expr, lower_bound: float, upper_bound: float, 
		s: Optional[str] = None):
		pass

	@abstractmethod
	def frequency2d(
		self, ex1: Expr, lb1: float, hb1: float, ex2: Expr, 
		lb2: float, hb2: float, s: Optional[str] = None):
		pass


"""
Actions class for producing HTML output
"""

__all__ = ["HTMLActions"]

from endplay.dealer.actions.base import BaseActions
from endplay.types import Denom, Player, Vul
from io import StringIO
import endplay.stats as stats
import matplotlib.pyplot as plt

class HTMLActions(BaseActions):
	def __init__(self, deals, stream, board_numbers):
		super().__init__(deals, stream, board_numbers, "html")
#
	def print(self, *players):
		compass = '<div class="compass">'
		for player in Player.iter_order("NWES"):
			compass += f'<div class="{player.name}">{player.abbr}</div>'
		compass += "</div>"
		box = "<div></div>"
		self.write("<div>")
		for i, deal in enumerate(self.deals, 1):
			if self.board_numbers:
				info = f'<div class="board-info"><div>Board {i}</div>'
				info += f'<div>Vul {Vul.from_board(i).abbr}</div><div>'
				info += f'{Player((i-1) % 4).abbr} deals</div></div>'
			else:
				info = box
			hands = {}
			for player in Player:
				if player in players:
					h = f'<div class="hand {player.name}">'
					for denom in Denom.suits():
						holding = str(deal[player][denom]) or "&mdash;"
						h += f'<div class="{denom.name}">{holding}</div>'
					hands[player] = h + "</div>"
				else:
					hands[player] = r'<div class="hand {player.name}"></div>'
			self.write('<div class="deal" style="margin: 20px">')
			self.write(
				info, hands[Player.north], box, 
				hands[Player.west], compass, hands[Player.east],
				box, hands[Player.south], box)
			self.write('</div>')
		self.write("</div>")

	def printpbn(self):
		self.write('<table class="deal-oneline">')
		self.write('<tr>', end='')
		if self.board_numbers:
			self.write('<th>#</th>', end='')
		self.write('<th>PBN</th>', end='')
		self.write("</tr>")
		for i, deal in enumerate(self.deals, 1):
			self.write('<tr>', end='')
			if self.board_numbers:
				self.write(f'<td><div class="value">{i}</div></td>', end='')
			self.write(f'<td><div class="hand-inline">{str(deal)}</div></td>')
			self.write('</tr>')
		self.write('</table>')

	def printcompact(self, expr = None):
		self.write('<table class="deal-compact">')
		self.write('<tr>')
		if self.board_numbers:
			self.write('<th>#</th>')
		self.write('<th>North</th><th>East</th><th>South</th><th>West</th>', end='')
		if expr is not None:
			self.write('<th>Value</th>', end='')
		self.write("</tr>")
		for i, deal in enumerate(self.deals, 1):
			self.write("<tr>")
			if self.board_numbers:
				self.write(f'<td><div class="value">{i}</div></td>', end='')
			for player in Player:
				self.write('<td><div class="hand">')
				for denom in Denom.suits():
					holding = str(deal[player][denom]) or "&mdash;"
					self.write(f'<div class="{denom.name}">{holding}</div>')
				self.write('</div></td>')
			if expr is not None:
				self.write(f'<td><div class="value">{expr(deal)}</div></td>')
			self.write('</tr>')
		self.write('</table>')
		
	def printoneline(self, expr = None):
		self.write('<table class="deal-oneline">')
		self.write('<tr>', end='')
		if self.board_numbers:
			self.write('<th>#</th>', end='')
		self.write('<th>North</th><th>East</th><th>South</th><th>West</th>', end='')
		if expr is not None:
			self.write('<th>Value</th>', end='')
		self.write("</tr>")
		for i, deal in enumerate(self.deals, 1):
			self.write("<tr>")
			if self.board_numbers:
				self.write(f'<td><div class="value">{i}</div></td>', end='')
			for player in Player:
				self.write('<td><div class="hand-inline">')
				for denom in Denom.suits():
					holding = str(deal[player][denom]) or "&mdash;"
					self.write(f'<div class="{denom.name} inline">{holding}</div>')
				self.write('</div></td>')
			if expr is not None:
				self.write(f'<td><div class="value">{expr(deal)}</div></td>')
			self.write('</tr>')
		self.write('</table>')

	def printes(self, *objs):
		self.write('<table class="deal-oneline">')
		self.write('<tr>', end='')
		if self.board_numbers:
			self.write('<th>#</th>', end='')
		self.write('<th>Output</th>', end='')
		self.write("</tr>")
		for i, deal in enumerate(self.deals, 1):
			self.write('<tr>', end='')
			if self.board_numbers:
				self.write(f'<td><div class="value">{i}</div></td>', end='')
			self.write('<td><div class="value">')
			for obj in objs:
				if callable(obj):
					self.write(obj(deal), end='')
				else:
					self.write(str(obj).replace('\n', "<br>"), end='')
			self.write('</div></td>')
			self.write('</tr>')
		self.write('</table>')

	def average(self, expr, s = None):
		if s is None:
			s = ""
		self.write('<div class="valuebox">', s, stats.average(self.deals, expr), "</div>")

	def frequency1d(self, expr, lb, ub, s = None):
		hist = stats.frequency(self.deals, expr, lb, ub)
		fig, ax = plt.subplots()
		ax.bar(list(range(lb, ub+1)), hist)
		if s: ax.set_title(s)
		f = StringIO()
		fig.savefig(f, format="svg")
		self.write('<div class="valuebox">', f.getvalue(), "</div>")

	def frequency2d(self, ex1, lb1, ub1, ex2, lb2, ub2, s = None):
		hist = stats.cofrequency(self.deals, ex1, ex2, lb1, ub1, lb2, ub2)
		fig, ax = plt.subplots()
		m = ax.matshow(hist)
		fig.colorbar(m)
		ax.set_xticks(list(range(1 + ub2 - lb2)))
		ax.set_xticklabels([str(i) for i in range(lb2, ub2+1)])
		ax.set_yticks(list(range(1 + ub1 - lb1)))
		ax.set_yticklabels([str(i) for i in range(lb1, ub1+1)])
		if s: ax.set_title(s)
		f = StringIO()
		fig.savefig(f, format="svg")
		self.write('<div class="valuebox">', f.getvalue(), "</div>")


"""
Parser for Dealer scripts
"""

__all__ = [ "DealerParser", "ParseException" ]

from typing import TextIO
from endplay.types import Player, Vul, Hand, Contract, Denom, Deal
import pyparsing as pp
pp.ParserElement.enablePackrat()
ParseException = pp.ParseException

DEBUG = False

def constructor(f):
	"""
	The parse actions are defined in the :class:`Node` class's namespace, and are 
	thus static methods. This decorator is simply a wrapper around `staticmethod`
	which can print out optional debugging information if `DEBUG` is set to `True`
	"""
	global DEBUG
	if DEBUG:
		def wrapper(string, location, tokens):
			try:
				res = f(string, location, tokens)
				print(tokens, "->", res)
				return res
			except Exception as e:
				print("Error construction AST:", e)

		return staticmethod(wrapper)
	else:
		return staticmethod(f)

class Node:
	"""
	Basic Node element from which the syntax tree is built up. Each node contains
	a value and dtype which describes how the value should be interpreted
	"""
	ROOT = 0 # Defines the root of the tree. ROOT notes should have no parents
	SYMBOL = 1  # Defines a symbol name which requires lookup at runtime
	OPERATOR = 2 # Defines a unary or binary operator with its arguments as children
	FUNCTION = 4 # Defines a function with its arguments as children
	ACTION = 5 # Defines an action with the optioal arguments as children
	VALUE = 6 # Defines a literal value. The value is stored as the corresponding Python type
	
	def __init__(self, value, dtype):
		self.value = value
		self.dtype = dtype
		self.children = []
		
	def append_child(self, child):
		self.children.append(child)

	@property
	def first_child(self) -> 'Node':
		return self.children[0]

	@property
	def middle_child(self) -> 'Node':
		return self.children[1]

	@property
	def last_child(self) -> 'Node':
		return self.children[-1]

	@property
	def n_children(self) -> int:
		return len(self.children)
		
	def pprint(self, indent = 0):
		print(f"{' ': >{indent}}↪ {self.value!r}", end='')
		if self.dtype == Node.VALUE: print(f" ({type (self.value)})")
		else: print(f" (dtype {self.dtype})")
		for child in self.children:
			try:
				child.pprint(indent + 2)
			except AttributeError:
				raise RuntimeError(f"Non-node child {child} found")
			
	def __repr__(self):
		return "{" + str(self.value) + (": " + ",".join(str(s) for s in self.children) if self.children else "") + "}"

	@constructor
	def from_symbol(string, location, tokens):
		return Node(tokens[0], Node.SYMBOL)

	@constructor
	def from_unaryop(string, location, tokens):
		args = tokens[0]
		op, arg = args[0], args[1]
		node = Node(op, Node.OPERATOR)
		node.append_child(arg)
		return node
	
	@constructor
	def from_binaryop(string, location, tokens):
		args = tokens[0]
		while len(args) > 1:
			lhs, op, rhs, *rest = args
			node = Node(op, Node.OPERATOR)
			node.append_child(lhs)
			node.append_child(rhs)
			args = [node, *rest]
		return args[0]
		
	@constructor
	def from_ternaryop(string, location, tokens):
		args = tokens[0]
		node = Node("if", Node.FUNCTION)
		node.append_child(args[0])
		node.append_child(args[2])
		node.append_child(args[4])
		return node

	@constructor
	def from_function(string, location, tokens):
		name, *args = tokens
		f = Node(name, Node.FUNCTION)
		for arg in args:
			f.append_child(arg)
		return f
		
	@constructor
	def from_input(string, location, tokens):
		f, *args = tokens
		node = Node(f, Node.ACTION)
		for arg in args:
			node.append_child(arg)
		return node
		
	@constructor
	def from_number(string, location, tokens):
		i, f = int(tokens[0]), float(tokens[0])
		return Node(i if i == f else f, Node.VALUE)
		
	@constructor
	def from_string(string, location, tokens):
		return Node(tokens[0][1:-1], Node.VALUE)
		
	@constructor
	def from_compass(string, location, tokens):
		player = Player.find(tokens[0])
		return Node(player, Node.VALUE)
		
	@constructor
	def from_vul(string, location, tokens):
		vul = Vul.find(tokens[0])
		return Node(vul, Node.VALUE)
		
	@constructor
	def from_suitholding(string, location, tokens):
		node = Node(tokens[0], Node.SUITHOLDING)
		holding = Node(tokens[1], Node.STRING)
		node.append_child(holding)
		return node
	
	@constructor
	def from_variable(string, location, tokens):
		node = Node("define", Node.ACTION)
		node.append_child(Node(tokens[0], Node.SYMBOL))
		node.append_child(tokens[2])
		return node
		
	@constructor
	def from_action(string, location, tokens):
		node = Node(tokens[0], Node.ACTION)
		for arg in tokens[1:]:
			if isinstance(arg, pp.ParseResults):
				for elem in arg:
					node.append_child(elem)
			else:
				node.append_child(arg)
		return node

	@constructor
	def from_contract(string, location, tokens):
		level = tokens[1]
		denom = tokens[2]
		contract = Contract(level=level, denom=denom)
		return Node(contract, Node.VALUE)

	@constructor
	def from_hand(string, location, tokens):
		hand = Hand()
		for token in tokens:
			denom = Denom.find(token[0])
			hand[denom] = token[1]
		return Node(hand, Node.VALUE)

	@constructor
	def from_predeal(string, location, tokens):
		deal = Deal()
		for i in range(1, len(tokens), 2):
			deal[tokens[i].value] = tokens[i+1].value
		action = Node("predeal", Node.ACTION)
		action.append_child(Node(deal, Node.VALUE))
		return action

	@constructor
	def from_pattern(string, location, tokens):
		shape = []
		for char in tokens[0]:
			if char in "xX":
				shape.append(None)
			else:
				shape.append(int(char))
		return Node(tuple(shape), Node.VALUE)

	@constructor
	def from_shape_any(string, location, tokens):
		node = Node("any", Node.OPERATOR)
		node.append_child(tokens[0][1])
		return node

	@constructor
	def from_shape_combine(string, location, tokens):
		args = tokens[0]
		while len(args) > 1:
			lhs, op, rhs, *rest = args
			node = Node(op, Node.OPERATOR)
			node.append_child(lhs)
			node.append_child(rhs)
			args = [node, *rest]
		return args[0]

	@constructor
	def from_shapelist(string, location, tokens):
		return tokens[0]

	@constructor
	def from_denom(string, location, tokens):
		denom = Denom.find(tokens[0])
		return Node(denom, Node.VALUE)

	@constructor
	def from_nl(string, location, tokens):
		return Node("\n", Node.VALUE)

	@constructor
	def from_card(string, location, tokens):
		return Node(f"{tokens[1]}{tokens[0]}", Node.VALUE)


def new_func(name, *args):
	if isinstance(name, str):
		name = pp.CaselessKeyword(name)
	name -= pp.Suppress("(")
	if len(args) != 0:
		for arg in args[:-1]:
			name += arg + pp.Suppress(",")
		name += args[-1]
	name += pp.Suppress(")")
	name.setParseAction(Node.from_function)
	return name

class DealerParser:
	def __init__(self):
		# Initialise most of the value types
		ppc = pp.pyparsing_common
		number = ppc.number
		number.setParseAction(Node.from_number)
		newline = pp.Keyword("\\n")
		newline.setParseAction(Node.from_nl)
		string = pp.dblQuotedString
		string.setParseAction(Node.from_string)
		compass = pp.oneOf("north south east west", caseless=True, asKeyword=True)
		compass.setParseAction(Node.from_compass)
		suit = pp.Regex(r"spades?") | pp.Regex(r"hearts?") | pp.Regex(r"diamonds?") | pp.Regex(r"clubs?")
		suit.setParseAction(Node.from_denom)
		strain = pp.Regex(r"notrumps?") | pp.Regex(r"spades?") | pp.Regex(r"hearts?") | pp.Regex(r"diamonds?") | pp.Regex(r"clubs?")
		strain.setParseAction(Node.from_denom)
		vul = pp.oneOf("none ns ew all", caseless=True, asKeyword=True)
		vul.setParseAction(Node.from_vul)
		card_suit = pp.oneOf("s h d c", caseless=True)
		card_rank = pp.Word("AKQJTakqjt98765432")
		card = pp.Word("AKQJTakqjt98765432") + pp.Word("SHDCshdc")
		card.setParseAction(Node.from_card)
		suitholding = pp.Group(card_suit + pp.OneOrMore(card_rank))
		hand = pp.delimitedList(suitholding)
		hand.setParseAction(Node.from_hand)
		pattern = pp.Word("0123456789xX", exact=4)
		pattern.setParseAction(Node.from_pattern)
		shapelist = pp.infixNotation(pattern, [
			(pp.oneOf("any"), 1, pp.opAssoc.RIGHT, Node.from_shape_any),
			(pp.oneOf("+ -"), 2, pp.opAssoc.LEFT, Node.from_shape_combine)
		])
		contract = pp.Literal("x") + pp.oneOf("1 2 3 4 5 6 7") + pp.oneOf("N S H D C")
		contract.setParseAction(Node.from_contract)

		# Functions in expressions. This does function name checking at parse time, which in my
		# first attempt at implementing this was the only way to make the grammar unambiguous. 
		# This should probably be replaced with something simpler.
		shape = new_func("shape", compass, shapelist)
		hascard = new_func("hascard", compass, card)
		suitlength = new_func(pp.Regex(r"spades?") | pp.Regex(r"hearts?") | pp.Regex(r"diamonds?") | pp.Regex(r"clubs?"), compass)
		hcp = new_func(pp.Regex("hcps?"), compass, suit) | new_func(pp.Regex("hcps?"), compass)
		ptN = new_func(pp.Regex("pt[0-9]"), compass) | new_func(pp.Regex("pt[0-9]"), compass, suit)
		for name in ["tens", pp.Regex("jacks?"), pp.Regex("queens?"), pp.Regex("kings?"), pp.Regex("aces?"), "top2", "top3", "top4", "top5", "c13"]:
			ptN |= new_func(name, compass) | new_func(name, compass, suit)
		control = new_func(pp.Regex("controls?"), compass) | new_func(pp.Regex("controls?"), compass, suit)
		loser = new_func(pp.Regex("losers?"), compass) | new_func(pp.Regex("losers?"), compass, suit)
		quality = new_func("cccc", compass) | new_func("quality", compass, suit)
		trick = new_func(pp.Regex("tricks?"), compass, strain)
		score = new_func("score", vul, contract, ppc.integer)
		imp = new_func(pp.Regex("imps?"), ppc.integer)
		func = shape | hascard | suitlength | hcp | ptN | control | loser | quality | trick | score | imp

		# Expressions (for conditions and variable definitions)
		expr = pp.Forward()
		symbol = pp.Regex(r"(?!(printall)|(print)|(printew)|(printpbn)|(printcompact)|(printes)|" + \
			r"(printoneline)|(average)|(frequency))[a-zA-Z0-9_-]+")
		symbol.setParseAction(Node.from_symbol)
		unary1 = pp.oneOf("! not")
		bin1 = pp.oneOf("* / %")
		bin2 = pp.oneOf("+ -")
		bin3 = pp.oneOf("<= < >= >")
		bin4 = pp.oneOf("== !=")
		bin5 = pp.oneOf("&& and")
		bin6 = pp.oneOf("|| or")
		ternary = ("?", ":")
		operator = pp.infixNotation(func | number | symbol, [
			(unary1, 1, pp.opAssoc.RIGHT, Node.from_unaryop),
			(bin1, 2, pp.opAssoc.LEFT, Node.from_binaryop),
			(bin2, 2, pp.opAssoc.LEFT, Node.from_binaryop),
			(bin3, 2, pp.opAssoc.LEFT, Node.from_binaryop),
			(bin4, 2, pp.opAssoc.LEFT, Node.from_binaryop),
			(bin5, 2, pp.opAssoc.LEFT, Node.from_binaryop),
			(bin6, 2, pp.opAssoc.LEFT, Node.from_binaryop),
			(ternary, 3, pp.opAssoc.RIGHT, Node.from_ternaryop)])		
		expr <<= operator
		
		# Actions
		printall = pp.CaselessKeyword("printall")
		printall.setParseAction(Node.from_action)
		print_compass = pp.CaselessKeyword("print") + pp.Suppress("(") - pp.Group(pp.delimitedList(compass)) + pp.Suppress(")")
		print_compass.setParseAction(Node.from_action)
		printew = pp.CaselessKeyword("printew")
		printew.setParseAction(Node.from_action)
		printpbn = pp.CaselessKeyword("printpbn")
		printpbn.setParseAction(Node.from_action)
		printcompact = pp.CaselessKeyword("printcompact") + pp.Optional(expr)
		printcompact.setParseAction(Node.from_action)
		printoneline = pp.CaselessKeyword("printoneline") + pp.Optional(expr)
		printoneline.setParseAction(Node.from_action)
		printes = pp.CaselessKeyword("printes") + pp.delimitedList(string | expr | newline)
		printes.setParseAction(Node.from_action)
		average = pp.CaselessKeyword("average") + pp.Optional(string) + expr
		average.setParseAction(Node.from_action)
		frequency = pp.CaselessKeyword("frequency") + pp.Optional(string) + pp.Suppress("(") + expr + \
			pp.Suppress(",") + ppc.number + pp.Suppress(",") + ppc.number + pp.Suppress(")")
		frequency.setParseAction(Node.from_action)
		frequency2 = pp.CaselessKeyword("frequency") + pp.Optional(string) + pp.Suppress("(") + expr + \
			pp.Suppress(",") + ppc.number + pp.Suppress(",") + ppc.number + pp.Suppress(",") + expr + \
			pp.Suppress(",") + ppc.number + pp.Suppress(",") + ppc.number + pp.Suppress(")")
		frequency2.setParseAction(Node.from_action)
		
		# Inputs
		generate = pp.CaselessKeyword("generate") + ppc.number
		generate.setParseAction(Node.from_input)
		produce = pp.CaselessKeyword("produce") + ppc.number
		produce.setParseAction(Node.from_input)
		vulnerable = pp.CaselessKeyword("vulnerable") + vul
		vulnerable.setParseAction(Node.from_input)
		dealer = pp.CaselessKeyword("dealer") + compass
		dealer.setParseAction(Node.from_input)
		predeal = pp.CaselessKeyword("predeal") + pp.OneOrMore(compass + hand)
		predeal.setParseAction(Node.from_predeal)
		pointcount = pp.CaselessKeyword("pointcount") + pp.delimitedList(ppc.number)
		pointcount.setParseAction(Node.from_input)
		altcount = pp.CaselessKeyword("altcount") + ppc.number + pp.delimitedList(ppc.number)
		altcount.setParseAction(Node.from_input)
		condition = pp.CaselessKeyword("condition") + expr
		condition.setParseAction(Node.from_input)
		action = pp.CaselessKeyword("action") + pp.delimitedList(
			printall | print_compass | printew | printpbn | printcompact |
			printoneline | printes | average | frequency | frequency2
		)
		action.setParseAction(Node.from_input)
		variable = ppc.identifier + pp.Literal("=") + expr
		variable.setParseAction(Node.from_variable)
		
		# Combine into the grammar for the whole file
		self.grammar = pp.ZeroOrMore(
			generate | produce | vulnerable | dealer | predeal | 
			pointcount | altcount | condition | action | variable
		)
		self.grammar.ignore(pp.cppStyleComment)
		# Note: from the docs I gather that we should only support #comment style
		# if the hash is at the beginning of the line, so this ignore statement
		# should be:
		#   self.grammar.ignore(pp.LineStart() + pp.pythonStyleComment)
		# but the test case fails, and as hashtags can't legally (probably?) appear
		# anywhere else I don't think it will break anything to be looser with the
		# requirements
		self.grammar.ignore(pp.pythonStyleComment)

		self.expr = expr
		
	def _build_tree(self, parseResults: pp.ParseResults) -> Node:
		root = Node("root", Node.ROOT)
		for action in parseResults:
			root.append_child(action)
		return root

	def parse_expr(self, s: str) -> Node:
		"""
		Parse an expression string into a syntax tree, for instance to compute
		the tree of a particular condition

		:param s: The condition string, e.g. "hcp(n) == 10 && shape(s) == 4432"
		:return: The root node of the syntax tree
		"""
		parseResults = self.expr.parseString(s, parseAll=True)
		return parseResults[0]

	def parse_file(self, f: TextIO) -> Node:
		"""
		Parse a file into a syntax tree

		:param f: A handle to a file or TextIO stream to be parsed
		:return: The root node of the syntax tree
		"""
		parseResults = self.grammar.parseFile(f, parseAll=True)
		return self._build_tree(parseResults)
		
	def parse_string(self, s: str) -> Node:
		"""
		Parse a strimg into a syntax tree

		:param s: The string to parse
		:return: The root node of the syntax tree
		"""
		parseResults = self.grammar.parseString(s, parseAll=True)
		return self._build_tree(parseResults)



"""
A shell-like interface to many of the features of endplay including 
dealing hands and performing double-dummy analysis.
"""

__all__ = ["CmdFrontend"]

import cmd
from endplay.interact import InteractiveDeal
from endplay.dealer import generate_deal
from endplay.types import Card, Deal, Vul, Player, Denom, Hand
from endplay.dds import solve_board, calc_dd_table, analyse_play, par
from endplay.evaluate import hcp
import traceback

class CmdFrontend(cmd.Cmd):
	def __init__(self, deal: InteractiveDeal, verbose_errors=False):
		self.deal = deal
		self.board_no = 0
		self.verbose_errors = verbose_errors
		self.needs_printing = True
		super().__init__()

	def interact(self):
		self.cmdloop()
		
	def onecmd(self, line):
		try:
			super().onecmd(line)
		except Exception as e:
			print("Encountered an unexpected error while executing:", e)
			if self.verbose_errors:
				traceback.print_exc()
		
	def cmdloop(self, intro=None):
		self.postcmd(False, "")
		try:
			super().cmdloop(intro)
		except KeyboardInterrupt:
			return

	def emptyline(self):
		pass
		
	def postcmd(self, stop, line):
		if self.needs_printing:
			print()
			self.deal.pprint(self.board_no or None)
			self.needs_printing = False
		self.prompt = self.deal.first.abbr + "> "
		return stop
		
	def do_ddtable(self, arg):
		"Calculate the double dummy table for the current deal"
		if len(self.deal.curtrick) != 0:
			print("No cards must be played to the current trick for this action")
			return
		table = calc_dd_table(self.deal)
		table.pprint()
		
	def do_play(self, arg):
		"""
		Play the listed cards.
		Example 1: play SA
		Example 2: play HQ S5 H2
		"""
		args = arg.split()
		for a in args:
			try:
				card = Card(a)
				self.deal.play(card)
			except ValueError:
				print(f"Could not parse `{a}` as a card name")
		self.needs_printing = True

	def do_reset(self, arg):
		"Reset the deal to its original state"
		self.deal.reset()
		self.needs_printing = True
		
	def do_checkpoint(self, arg):
		"Erase the undo history so that reset will take you back to this point"
		self.deal.checkpoint()
		
	def do_set(self, arg):
		"""
		Set a hand to given PBN string
		Example: set S AQ85.532..JT92
		No error checking for duplicate cards or incorrect number of cards is performed.
		"""
		if self.deal.curtrick:
			print("No cards must be played to the current trick for this action")
			return
		args = arg.split()
		if len(args) != 2:
			print(f"Expected two arguments (player and PBN string), got {len(args)}")
		try:
			player = Player.find(args[0])
		except ValueError:
			print(f"Could not parse {args[0]} as player name")
			return
		try:
			self.deal[player] = Hand(args[1])
		except RuntimeError:
			print(f"Invalid PBN string: `{args[0]}`")
		self.needs_printing = True

	def do_solve(self, arg):
		"""
		Display the double-dummy maximum number of tricks playing each card in the current
		player's hand can yield.
		"""
		res = {}
		for card, tricks in solve_board(self.deal):
			if tricks in res:
				res[tricks].append(card)
			else:
				res[tricks] = [card]
		for tricks, cards in res.items():
			print(f"{tricks}:", "  ".join(str(c) for c in cards))
		
	def do_redeal(self, arg):
		"""
		Redeal the hand to the given PBN string (or an empty string for a blank deal)
		Example 1: redeal
		Example 2: redeal N:95..A. 8.5.Q. .QT5.. Q..T4.
		"""
		self.deal.reset()
		if arg:
			try:
				new_deal = Deal(arg)
				for player, hand in new_deal:
					self.deal[player] = hand
			except RuntimeError:
				print(f"Invalid PBN string: `{arg}`")
		self.needs_printing = True

	def do_shuffle(self, arg):
		"""
		Generate a random deal satisfying the given constraints.
		Example 1: shuffle
		Example 2: shuffle hcp(north) > hcp(south)
		"""
		if arg:
			try:
				new_deal = generate_deal(arg)
			except RuntimeError:
				raise RuntimeError("Could not generate deal satisfying this constraint")
		else:
			new_deal = generate_deal()
		self.deal.reset()
		for player, hand in new_deal:
			self.deal[player] = hand
		self.needs_printing = True

	def do_first(self, arg):
		"""
		If no argument is provided, display the position of the player to play the first card to
		the current trick. If a position is given, set that seat to be on lead
		Example 1: first
		Example 2: first W
		"""
		if not arg:
			print(self.deal.first.name)
			return
			
		if len(self.deal.curtrick) != 0:
			print("No cards must be played to the current trick for this action")
			return
		try:
			player = Player.find(arg)
			self.deal.first = player
		except ValueError:
			print(f"Unrecognised player: {arg}")

	def do_board(self, arg):
		"""
		If no argument is provided, display the board number. If a number is given, set the
		board number to this. Set to 0 to disable the board number.
		Example 1: board
		Example 2: board 12
		"""
		if not arg:
			print(self.board_no)
		else:
			self.board_no = int(arg)
			self.needs_printing = True
		
	def do_trump(self, arg):
		"""
		If no argument is provided, display the trump suit of the current deal. If a suit is
		given, set the trump suit to this.
		Example 1: trump
		Example 2: trump NT
		"""
		if not arg:
			print(self.deal.trump.name)
			return
			
		if len(self.deal.curtrick) != 0:
			print("No cards must be played to the current trick for this action")
			return
		denom = Denom.find(arg)
		self.deal.trump = denom

	def do_hcp(self, arg):
		player = Player.find(arg)
		print(hcp(self.deal[player]))
		
	def do_analyse(self, arg):
		"""
		If no arguments are provided, display the double dummy results of playing each card
		in the hand currently on lead. If a list of cards are provided, display the double
		dummy results for each card as it is played
		"""
		play = [Card(c) for c in arg.split()]
		first, *rest = analyse_play(self.deal, play)
		print("  ", first)
		for card, tricks in zip(play, rest):
			print(card, tricks)

	def do_par(self, arg):
		"""
		Calculate the par contract and score for the board using the first player as the
		dealer
		"""
		vul = Vul.from_board(self.board_no)
		p = par(self.deal, vul, self.deal.first)
		print(f"Par score: {p.score}")
		for contract in p:
			print(" -", contract)

	def do_display(self, arg):
		"Displays the current deal"
		self.needs_printing = True

	def do_pbn(self, arg):
		"Displays the current deal as a PBN string"
		print(str(self.deal))
		
	def do_undo(self, arg):
		"Undo the previous action, if it changed the current deal"
		if not self.deal.undo():
			print("Nothing to undo")
		else:
			self.needs_printing = True

	def do_history(self, arg):
		"""Print the undo history. Accepts an integer with the maximum number
		of actions to show
		"""
		if arg:
			n = int(arg)
			for action in self.deal._history[:-n]:
				print(action)
		else:
			for action in self.deal._history:
				print(action)
			
	def do_exit(self, arg):
		"Exits the interaction"
		raise KeyboardInterrupt



"""
Functions for escaping suit symbols
"""

from __future__ import annotations

__all__ = ["escape_suits", "unescape_suits"]

import re

_escape_suits_table = str.maketrans({
	"♠": "!S",
	"♥": "!H",
	"♦": "!D",
	"♣": "!C"
})

def escape_suits(s: str):
	"""
	Escape unicode suit symbols into BBO suit notation (!S, !H, !D, !C)
	"""
	return s.translate(_escape_suits_table)

def unescape_suits(s: str):
	"""
	Unescape BBO suit notation (!S, !H, !D, !C) into unicode suit symbols
	"""
	s = re.sub("!s", "♠", s, flags=re.IGNORECASE)
	s = re.sub("!h", "♥", s, flags=re.IGNORECASE)
	s = re.sub("!d", "♦", s, flags=re.IGNORECASE)
	return re.sub("!c", "♣", s, flags=re.IGNORECASE)



#!/usr/bin/env python3

import os
import pathlib
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

class CMakeExtension(Extension):
	"""
	Stub class to distinguish between default extensions and CMake
	extensions (which contain no sources as these are listed in the
	CMakeLists.txt file)
	"""
	def __init__(self, name):
		# don't invoke the original build_ext for this special extension
		super().__init__(name, sources=[])

class cmakeable_build_ext(build_ext):
	"""
	build_ext compatible class which detects if the extension it is to
	build is a CMakeExtension in which case it delegates building to
	the CMake executable.
	"""
	def run(self):
		for ext in self.extensions:
			if isinstance(ext, CMakeExtension):
				self.build_cmake(ext)
		super().run()

	def build_cmake(self, ext):
		cwd = pathlib.Path().absolute()

		# Create directory structure
		build_temp = pathlib.Path(self.build_temp)
		build_temp.mkdir(parents=True, exist_ok=True)
		extdir = pathlib.Path(self.get_ext_fullpath(ext.name))
		extdir.mkdir(parents=True, exist_ok=True)

		# Check which architecture we should be building for
		import struct
		bits = struct.calcsize('P') * 8

		# Setup args passed to cmake
		config = 'Debug' if self.debug else 'Release'
		cmake_config_args = [
			'-DCMAKE_INSTALL_PREFIX=' + str(extdir.parent.absolute()),
			'-DCMAKE_BUILD_TYPE=' + config,
			'-DSETUPTOOLS_BUILD=1'
		]
		if os.name == 'nt':
			if bits == 64: cmake_config_args.append('-A x64')
			elif bits == 32: cmake_config_args.append('-A Win32')
			else: raise RuntimeError(f"Unknown computer architecture with {bits} bits")
		else:
			if bits == 32: cmake_config_args.append('-DCOMPILE_32_BITS=1')

		cmake_build_args = [
			"--build", ".",
			"--target", "install",
			"--config", config
		]

		# Disable warning MSB8029 (https://stackoverflow.com/a/60301902/5194459)
		os.environ["IgnoreWarnIntDirInTempDetected"] = "true"

		os.chdir(str(build_temp))
		self.spawn(['cmake', str(cwd)] + cmake_config_args)
		if not self.dry_run:
			self.spawn(['cmake'] + cmake_build_args)
		os.chdir(str(cwd))

setup(
	ext_modules = [CMakeExtension('endplay')],
	cmdclass = {
		'build_ext': cmakeable_build_ext,
	},
	test_suite = "tests"
)


"""
The endplay.dealer program is a reimplementation of Hans van Staveren's dealer program
which can be used to generate hands for partnerships bidding training or for
generating statistics that can be used to design conventions, or win postmortems.

While compatibility with the original has been a design goal, the precise format of the
output is not expected to correspond 1:1 with the original program. Furthermore, several
extra options including the ability to specify actions and constraints at the command line
and some extra flags for specifying output formats are included. 
"""

import argparse
import traceback
import warnings
from endplay.dealer.runscript import run_script
import endplay.config as config

def main():
	# Parse command line
	parser = argparse.ArgumentParser(description=__doc__)
	deprecated = parser.add_argument_group(title="Deprecated options", description=
		"The following flags are listed as they were flags for the original dealer program"+\
		", but are now either ignored or cause the program to exit with an error")
	deprecated.add_argument("-e", action="store_true", help="Exhaust mode (alpha version).")
	parser.add_argument("-g", type=int, default=1000000, 
		help="Maximum number of hands to generate (default 1000000).")
	deprecated.add_argument("-l", type=int, 
		help="Instead of shuffling, deals are read from the file library.dat by M. Ginsberg")
	parser.add_argument("-m", action="store_true", help="Shows a progress meter.")
	parser.add_argument("-p", type=int, default=40, help="Maximum number of hands to produce (default 40).")
	deprecated.add_argument("-q", action="store_true", 
		help="Suppress PBN output (useful for testing, then switch it back on when generating the \"final\" sample).")
	parser.add_argument("-s", type=int, 
		help="Uses number as the seed for the random generator, running the program twice with the same seed, will produce the same sequence of hands.")
	parser.add_argument("-u", action="store_true", 
		help="Use unicode card symbols instead of the letters S, H, D, C")
	parser.add_argument("-v", action="count", default=0, 
		help="Verbose output, can be specified up to three times (-vvv). Level 1 prints statistics at the end of the run, level 2 prints any runtime generated warnings and level 3 prints full stacktraces if an exception occurs.")
	parser.add_argument("-V", action="store_true", help="Emits a version-identification string and exits")
	swapping = parser.add_mutually_exclusive_group()
	parser.set_defaults(swapping=0)
	swapping.add_argument("-0", action="store_const", dest="swapping", const=0, 
		help="No swapping, each deal is generated normally (shuffling), default.")
	swapping.add_argument("-2", action="store_const", dest="swapping", const=2,
		help="2-way swapping, after each shuffle another deal is generated by permuting E and W, leaving N and S in place (NB: not fully compatible with predeal).")
	swapping.add_argument("-3", action="store_const", dest="swapping", const=3,
		help="3-way swapping, after each shuffle another 5 deals are generated by permuting E, W, and S every which way, leaving N in place (NB: not fully compatible with predeal). not needed (NB: not compatible at all with predeal).")
	parser.add_argument("-c", action='append', default=[], 
		help="Constraint to be applied, can be specified more than once for multiple constraints")
	parser.add_argument("-f", choices=["plain", "latex", "pdf", "html"], help="Output format")
	parser.add_argument("-o", default=None, help="Name of file to save output to")
	parser.add_argument("-a", action='append', default=[], 
		help="Action to be applied, can be specified more than once")
	parser.add_argument("-b", action="store_true", help="Assign board numbers to each generated deal")
	parser.add_argument("-d", help="Cards to predeal, e.g. \"west S975,H64 east DA,C64\"")
	parser.add_argument("script", nargs='?', help="input file containing hand-descriptions and action")

	args = parser.parse_args()
	if args.V:
		exit(str(config.__version__))
	config.use_unicode = args.u
	

	# Warn/quit on deprecated flags
	if args.l:
		exit("Ginsberg library is not currently implemented")
	if args.e:
		print("Deprecated option -e ignored")
	if args.q:
		print("Deprecated option -q ignored")

	# Set up warning filter to only display warnings if verbosity >= 2
	if args.v < 2:
		warnings.simplefilter("ignore")
	else:
		warnings.showwarning = lambda msg, *args: print("Warning:", msg)

	# And away we go
	try:
		run_script(
			args.script, 
			show_progress = args.m, 
			produce = args.p, 
			generate = args.g, 
			seed = args.s, 
			verbose = args.v > 0, 
			swapping = args.swapping, 
			outformat = args.f, 
			outfile = args.o, 
			constraints = args.c, 
			actions = args.a, 
			predeal = args.d, 
			board_numbers = args.b)
	except Exception as e:
		print("dealer had to exit prematurely because of the following error:", e)
		if args.v >= 3:
			traceback.print_exc()
		exit(1)

if __name__ == "__main__":
	main()


"""
Functions for statistical analysis of deals.
"""

from __future__ import annotations

__all__ = ["average", "frequency", "cofrequency"]

from typing import Optional
from collections.abc import Iterable
from endplay.types import Deal
from endplay.dealer.constraint import Expr
from math import floor, ceil
try:
	from statistics import fmean
except ImportError:
	from statistics import mean as fmean


def average(deals: Iterable[Deal], func: Expr):
	"Return the average of a function over a sequence of deals"
	return fmean(func(deal) for deal in deals)

def frequency(deals: Iterable[Deal], func: Expr, lb: int, ub: int) -> list[int]:
	"""
	Calculate the value of a function over a range of deals, and bin the results into
	unit-sized bins around integer values from `lb` to `ub`
	
	:param deals: The input sequence of deals
	:param func: The function to evaluate over `deals`
	:param lb: Value below which values are ignored
	:param ub: Value above which values are ignored
	:return: An array of bins, and a tuple containing the left and right boundaries
	"""
	lb, ub = floor(lb), ceil(ub)
	bins = list(range(lb, ub+1))
	hist = [0] * len(bins)
	for deal in deals:
		val = round(func(deal))
		if bins[0] <= val <= bins[-1]:
			hist[val - bins[0]] += 1
	return hist

def cofrequency(
	deals: Iterable[Deal],
	func1: Expr,
	func2: Expr,
	lb1: Optional[float] = None,
	ub1: Optional[float] = None,
	lb2: Optional[float] = None,
	ub2: Optional[float] = None) -> list[list[int]]:
	"""
	Calculate the value of two functions over a range of deals, and bin the results into
	unit-sized bins around integer values from `lb` to `ub` to form a matrix

	:param deals: The input sequence of deals
	:param func1: The first function to evaluate over `deals`
	:param func2: The second function to evaluate over `deals`
	:param lower_bound1: Value below which values returned from func1 are ignored
	:param upper_bound1: Value above which values returned from func1 are ignored
	:param lower_bound2: Value below which values returned from func2 are ignored
	:param upper_bound2: Value above which values returned from func2 are ignored
	"""
	lb1, ub1 = floor(lb1), ceil(ub1)
	lb2, ub2 = floor(lb2), ceil(ub2)
	data1, data2 = [], []
	for deal in deals:
		data1.append(round(func1(deal)))
		data2.append(round(func2(deal)))
	bins1 = list(range(lb1, ub1+1))
	bins2 = list(range(lb2, ub2+1))
	hist = [[0 for _ in bins2] for _ in bins1]
	for val1 in data1:
		if not (lb1 <= val1 <= ub1):
			continue
		for val2 in data2:
			if not (lb2 <= val2 <= ub2):
				continue
			hist[val1 - bins1[0]][val2 - bins2[0]] += 1
	return hist


"""
Higher-level wrappers around Bo Haglund's dds wrapper, using the
Python types defined in `endplay.types` and which use sensible
defaults for the internal state such as threading indexes.
"""

__all__ = [
	'solve_board', 'solve_all_boards', 'calc_dd_table',
	'calc_all_tables', 'analyse_play', 'analyse_all_plays', 'par',
	'analyse_start', 'analyse_all_starts'
]

import endplay._dds as _dds
from endplay.dds.solve import solve_board, solve_all_boards
from endplay.dds.ddtable import calc_dd_table, calc_all_tables
from endplay.dds.analyse import analyse_play, analyse_all_plays, analyse_start, analyse_all_starts
from endplay.dds.parscore import par



"""
Collection of bridge-related types based on the binary format used by
Bo Haglund's DDS library but with a more Pythonic interface. These types
are used by all the submodules in the endplay library.
"""

__all__ = [
	"Vul", "Rank", "AlternateRank", "Player", "Denom", "Card",
	"Hand", "Deal", "Contract", "Penalty", "SuitHolding", "Bid",
	"Board", "ContractBid", "PenaltyBid"]

from endplay.types.vul import Vul
from endplay.types.rank import Rank, AlternateRank
from endplay.types.player import Player
from endplay.types.denom import Denom
from endplay.types.card import Card
from endplay.types.suitholding import SuitHolding
from endplay.types.hand import Hand
from endplay.types.deal import Deal
from endplay.types.contract import Contract
from endplay.types.penalty import Penalty
from endplay.types.bid import Bid, ContractBid, PenaltyBid
from endplay.types.board import Board



"""
Solving functions from the DDS library, which calculate the double dummy
results for playing each card in a player's hand.
"""

from __future__ import annotations

__all__ = ["SolvedBoard", "SolvedBoardList", "solve_board", "solve_all_boards"]

from typing import Optional
import endplay._dds as _dds
from endplay.types import Deal, Card, Denom, Rank
from collections.abc import Iterable, Iterator, Sequence

class SolvedBoard(Iterable):
	def __init__(self, data: _dds.solvedBoard):
		self._data = data

	def __iter__(self) -> Iterator[tuple[Card, int]]:
		":return: An iterator of (card, maxTrick) pairs"
		for i in range(self._data.cards):
			if self._data.score[i] == -1:
				continue
			holding = (1 << self._data.rank[i]) | self._data.equals[i]
			for rank in Rank:
				if holding & rank.value:
					card = Card(suit=Denom(self._data.suit[i]), rank=rank)
					yield (card, self._data.score[i])

	def __repr__(self) -> str:
		return '<SolvedBoard object>'

	def __str__(self) -> str:
		return '{' + ", ".join("b[0]:b[1]" for b in self) + '}'

class SolvedBoardList(Sequence):
	def __init__(self, data: _dds.solvedBoards):
			self._data = data

	def __len__(self) -> int:
		":return: The number of boards in the list"
		return self._data.noOfBoards

	def __getitem__(self, i: int) -> SolvedBoard:
		":return: The solved board at index `i`"
		if i < 0:
			i = len(self) - i
		if i < 0 or i >= len(self):
			return IndexError
		return SolvedBoard(self._data.solvedBoard[i])

	def __repr__(self) -> str:
		return f'<SolvedBoardList; length={len(self)}>'

	def __str__(self) -> str:
		if len(self) == 0:
			return "[]"
		return "[(" + ", ".join(str(b) for b in self) + ")]"

def solve_board(deal: Deal, target: Optional[int] = None) -> SolvedBoard:
	"""
	Calculate the double dummy score for all cards in the hand which is currently
	to play in a given deal

	:param deal: The deal to solve, with `first` and `trump` filled in
	:param target: If provided, only return cards which can make at least this many tricks
	"""
	fut = _dds.futureTricks()
	if target is None or target <= 0:
		target, solutions = -1, 3
	else:
		solutions = 2
	_dds.SolveBoard(deal._data, target, solutions, 0, fut, 0)
	return SolvedBoard(fut)

def solve_all_boards(deals: Iterable[Deal], target: Optional[int] = None) -> SolvedBoardList:
	"""
	Optimized version of solve_board for multiple deals which uses threading to
	speed up the calculation

	:param deals: The collection of boards to be solved, each with `first` and `trump` filled
	:param target: If provided, only return cards which can make at least this many tricks
	"""
	# Convert deals into boards
	bop = _dds.boards()
	bop.noOfBoards = 0
	if target is None or target <= 0:
		target, solutions = -1, 3
	else:
		solutions = 2
	for i, deal in enumerate(deals):
		if i > _dds.MAXNOOFBOARDS:
			raise RuntimeError(f"Too many boards, maximum is {_dds.MAXNOOFBOARDS}")
		bop.deals[i] = deal._data
		bop.target[i] = target
		bop.solutions[i] = solutions
		bop.mode[i] = 0
		bop.noOfBoards += 1

	solvedp = _dds.solvedBoards()
	_dds.SolveAllBoardsBin(bop, solvedp)
	return SolvedBoardList(solvedp)